from pydantic import BaseModel
from typing import Any


def format_field(name: str) -> str:
    return name.replace("_", " ").title()

def get_fields_for_model(costs: BaseModel) -> dict[str, Any]:
    cost_parts = {}
    for field_name, value in costs.model_dump().items():
        label = format_field(field_name)
        if isinstance(value, float):
            cost_parts[label] = value
        else:
            cost_parts[label] = value
    return cost_parts



import logging
import pytest

from gpt_token_tracker.pricing import TOKENS_PER_PRICE


# -----------------------------
# Tests
# -----------------------------

def test_compute_costs_realtime(token_logger_realtime, fake_realtime_usage, realtime_pricing):
    costs = token_logger_realtime.compute_costs(fake_realtime_usage)

    expected_text = (50 * realtime_pricing.pricing["text_in"]) / TOKENS_PER_PRICE
    expected_audio = (100 * realtime_pricing.pricing["audio_in"]) / TOKENS_PER_PRICE
    expected_image = (10 * realtime_pricing.pricing["image_in"]) / TOKENS_PER_PRICE
    expected_cached_text = (5 * realtime_pricing.pricing["cached_text_in"]) / TOKENS_PER_PRICE
    expected_cached_audio = 0
    expected_cached_image = 0
    expected_output_text = (30 * realtime_pricing.pricing["text_out"]) / TOKENS_PER_PRICE
    expected_output_audio = (20 * realtime_pricing.pricing["audio_out"]) / TOKENS_PER_PRICE

    expected_input_total = (
        expected_text
        + expected_audio
        + expected_image
        + expected_cached_text
        + expected_cached_audio
        + expected_cached_image
    )
    expected_output_total = expected_output_text + expected_output_audio
    expected_total = expected_input_total + expected_output_total

    assert costs.input_text_cost == pytest.approx(expected_text)
    assert costs.input_audio_cost == pytest.approx(expected_audio)
    assert costs.input_image_cost == pytest.approx(expected_image)

    assert costs.cached_text_cost == pytest.approx(expected_cached_text)
    assert costs.cached_audio_cost == expected_cached_audio
    assert costs.cached_image_cost == expected_cached_image

    assert costs.text_output_cost == pytest.approx(expected_output_text)
    assert costs.audio_output_cost == pytest.approx(expected_output_audio)

    assert costs.total_input_cost == pytest.approx(expected_input_total)
    assert costs.total_output_cost == pytest.approx(expected_output_total)
    assert costs.total_cost == pytest.approx(expected_total)


def test_record_logs_realtime(log_writer, realtime_model, realtime_result, token_logger_realtime, fake_realtime_usage, caplog):
    with caplog.at_level(logging.INFO):
        token_logger_realtime.record(realtime_model, realtime_result, fake_realtime_usage)

    assert len(caplog.records) == 1
    msg = caplog.records[0].message

    assert msg == "Model: gpt-realtime-mini | Result: The capital of Ireland is Dublin | Total Tokens: 215 | Input Tokens: 165 | Output Tokens: 50 | Input Tokens: 165 | Input Text Tokens: 50 | Input Audio Tokens: 100 | Input Image Tokens: 10 | Cached Text Tokens: 5 | Cached Audio Tokens: 0 | Cached Image Tokens: 0 | Output Tokens: 50 | Output Text Tokens: 30 | Output Audio Tokens: 20 | Input Text Cost: 0.000030 | Input Audio Cost: 0.001000 | Input Image Cost: 0.000050 | Cached Text Cost: 0.000000 | Cached Audio Cost: 0.000000 | Cached Image Cost: 0.000000 | Text Output Cost: 0.000072 | Audio Output Cost: 0.000400 | Total Input Cost: 0.001080 | Total Output Cost: 0.000472 | Total Cost: 0.001552"


def test_compute_costs_realtime_transcribe(token_logger_realtime_transcribe, fake_audio_transcription_usage, realtime_transcribe_pricing):
    costs = token_logger_realtime_transcribe.compute_costs(fake_audio_transcription_usage)

    expected_text_cost = (1 * realtime_transcribe_pricing.pricing["text_in"]) / TOKENS_PER_PRICE
    expected_audio_cost = (101 * realtime_transcribe_pricing.pricing["audio_in"]) / TOKENS_PER_PRICE
    expected_output_text = (7 * realtime_transcribe_pricing.pricing["text_out"]) / TOKENS_PER_PRICE

    expected_input_total = (
        expected_text_cost
        + expected_audio_cost
    )
    expected_total = expected_input_total + expected_output_text

    assert costs.input_text_cost == pytest.approx(expected_text_cost)
    assert costs.input_audio_cost == pytest.approx(expected_audio_cost)

    assert costs.output_text_cost == pytest.approx(expected_output_text)

    assert costs.total_cost == pytest.approx(expected_total)


def test_record_logs_realtime_transcribe(log_writer, transcribe_model, realtime_transcription_result, token_logger_realtime_transcribe, fake_audio_transcription_usage, caplog):
    with caplog.at_level(logging.INFO):
        token_logger_realtime_transcribe.record(transcribe_model, realtime_transcription_result, fake_audio_transcription_usage)

    assert len(caplog.records) == 1
    msg = caplog.records[0].message

    assert msg == "Model: gpt-4o-mini-transcribe | Result: What is the capital of Ireland? | Total Tokens: 109 | Input Tokens: 102 | Output Tokens: 7 | Input Text Tokens: 1 | Input Audio Tokens: 101 | Output Text Tokens: 7 | Input Text Cost: 0.000001 | Input Audio Cost: 0.001010 | Output Text Cost: 0.000017 | Total Cost: 0.001027"


def test_compute_costs_text_completion(
    token_logger_text_completion,
    fake_text_completion_usage,
    text_completion_pricing,
):
    costs = token_logger_text_completion.compute_costs(fake_text_completion_usage)

    expected_text = (165 * text_completion_pricing.pricing["text_in"]) / TOKENS_PER_PRICE
    expected_cached_text = (5 * text_completion_pricing.pricing["cached_text_in"]) / TOKENS_PER_PRICE
    expected_output_text = (50 * text_completion_pricing.pricing["text_out"]) / TOKENS_PER_PRICE

    expected_total = expected_text + expected_cached_text + expected_output_text

    assert costs.input_text_cost == pytest.approx(expected_text)
    assert costs.cached_text_cost == pytest.approx(expected_cached_text)
    assert costs.output_text_cost == pytest.approx(expected_output_text)
    assert costs.total_cost == pytest.approx(expected_total)

    assert costs.input_text_tokens == 165
    assert costs.cached_text_tokens == 5
    assert costs.output_text_tokens == 50


def test_record_logs_text_completion(log_writer, text_completion_model, text_completion_result, token_logger_text_completion, fake_text_completion_usage, caplog):
    with caplog.at_level(logging.INFO):
        token_logger_text_completion.record(text_completion_model, text_completion_result, fake_text_completion_usage)

    assert len(caplog.records) == 1
    msg = caplog.records[0].message

    assert msg == "Model: gpt-5-mini | Result: The capital of Ireland is Dublin | Total Tokens: 215 | Input Tokens: 165 | Output Tokens: 50 | Input Text Tokens: 165 | Cached Text Tokens: 5 | Output Text Tokens: 50 | Input Text Cost: 0.000041 | Cached Text Cost: 0.000000 | Output Text Cost: 0.000100 | Total Cost: 0.000141"


def test_close_is_noop(log_writer):
    # Should not raise
    log_writer.close()



from pydantic import BaseModel

from .writers.types import Writer, UsageTypes
from .pricing import BasePricing


class TokenLogger:
    """Routes usage objects to the appropriate writer. Writers compute costs."""

    def __init__(
        self,
        writer: Writer,
        pricing: BasePricing,
    ):
        self.pricing = pricing
        self.writer = writer
        self.writer.initialize(pricing.pricing)

    def compute_costs(self, usage: UsageTypes):
        return self.pricing.compute_costs(usage)

    def log(self, model: str, result: str, usage: UsageTypes, costs: BaseModel):
        self.writer.write(model, result, usage, costs)

    def record(self, model: str, result: str, usage: UsageTypes):
        costs = self.compute_costs(usage)
        self.log(model, result, usage, costs)

    def close(self):
        self.writer.close()



from pydantic import BaseModel


class RealtimeModel(BaseModel):
    input_tokens: int
    input_text_tokens: int
    input_audio_tokens: int
    input_image_tokens: int
    cached_text_tokens: int
    cached_audio_tokens: int
    cached_image_tokens: int
    output_tokens: int
    output_text_tokens: int
    output_audio_tokens: int
    input_text_cost: float
    input_audio_cost: float
    input_image_cost: float
    cached_text_cost: float
    cached_audio_cost: float
    cached_image_cost: float
    text_output_cost: float
    audio_output_cost: float
    total_input_cost: float
    total_output_cost: float
    total_cost: float


class RealtimeTranscribeModel(BaseModel):
    input_text_tokens: int
    input_audio_tokens: int
    output_text_tokens: int
    input_text_cost: float
    input_audio_cost: float
    output_text_cost: float
    total_cost: float


class ChatCompletionModel(BaseModel):
    input_text_tokens: int
    cached_text_tokens: int
    output_text_tokens: int
    input_text_cost: float
    cached_text_cost: float
    output_text_cost: float
    total_cost: float



from pathlib import Path
from pydantic import BaseModel
from typing import Protocol
from openai.types.responses.response_usage import ResponseUsage
from openai.types.realtime.realtime_response_usage import RealtimeResponseUsage
from openai.types.realtime.conversation_item_input_audio_transcription_completed_event import UsageTranscriptTextUsageTokens


UsageTypes = ResponseUsage | RealtimeResponseUsage | UsageTranscriptTextUsageTokens


class Writer(Protocol):
    """Writers must accept a str name or file path."""
    def __init__(self, name: str | Path): ...

    def initialize(self, pricing: dict[str, float]) -> None: ...

    def write(self, model: str, result: str, usage: UsageTypes, costs: BaseModel): ...

    def close(self): ...




from openai.types.responses.response_usage import ResponseUsage
from openai.types.realtime.realtime_response_usage import RealtimeResponseUsage
from openai.types.realtime.conversation_item_input_audio_transcription_completed_event import UsageTranscriptTextUsageTokens
from .models import RealtimeModel, ChatCompletionModel, RealtimeTranscribeModel
from .writers.types import UsageTypes

from abc import ABC, abstractmethod


TOKENS_PER_PRICE = 1_000_000


class BasePricing(ABC):
    def __init__(self, pricing: dict[str, float]):
        self.pricing = pricing

    @abstractmethod
    def compute_costs(self, u: UsageTypes):
        raise NotImplementedError


class PricingRealtime(BasePricing):
    def compute_costs(self, u: RealtimeResponseUsage) -> RealtimeModel:
        # Token counts
        input_text_tokens = u.input_token_details.text_tokens
        input_audio_tokens = u.input_token_details.audio_tokens
        input_image_tokens = u.input_token_details.image_tokens

        cached_text_tokens = u.input_token_details.cached_tokens_details.text_tokens
        cached_audio_tokens = u.input_token_details.cached_tokens_details.audio_tokens
        cached_image_tokens = u.input_token_details.cached_tokens_details.image_tokens

        output_text_tokens = u.output_token_details.text_tokens
        output_audio_tokens = u.output_token_details.audio_tokens

        input_tokens = (
            input_text_tokens
            + input_audio_tokens
            + input_image_tokens
            + cached_text_tokens
            + cached_audio_tokens
            + cached_image_tokens
        )

        output_tokens = output_text_tokens + output_audio_tokens

        # Costs
        input_text_cost = (input_text_tokens * self.pricing.get("text_in", 0)) / TOKENS_PER_PRICE
        input_audio_cost = (input_audio_tokens * self.pricing.get("audio_in", 0)) / TOKENS_PER_PRICE
        input_image_cost = (input_image_tokens * self.pricing.get("image_in", 0)) / TOKENS_PER_PRICE

        cached_text_cost = (cached_text_tokens * self.pricing.get("cached_text_in", 0)) / TOKENS_PER_PRICE
        cached_audio_cost = (cached_audio_tokens * self.pricing.get("cached_audio_in", 0)) / TOKENS_PER_PRICE
        cached_image_cost = (cached_image_tokens * self.pricing.get("cached_image_in", 0)) / TOKENS_PER_PRICE

        text_output_cost = (output_text_tokens * self.pricing.get("text_out", 0)) / TOKENS_PER_PRICE
        audio_output_cost = (output_audio_tokens * self.pricing.get("audio_out", 0)) / TOKENS_PER_PRICE

        total_input_cost = (
            input_text_cost
            + input_audio_cost
            + input_image_cost
            + cached_text_cost
            + cached_audio_cost
            + cached_image_cost
        )

        total_output_cost = text_output_cost + audio_output_cost
        total_cost = total_input_cost + total_output_cost

        return RealtimeModel(
            input_tokens=input_tokens,
            input_text_tokens=input_text_tokens,
            input_audio_tokens=input_audio_tokens,
            input_image_tokens=input_image_tokens,

            cached_text_tokens=cached_text_tokens,
            cached_audio_tokens=cached_audio_tokens,
            cached_image_tokens=cached_image_tokens,

            output_tokens=output_tokens,
            output_text_tokens=output_text_tokens,
            output_audio_tokens=output_audio_tokens,

            input_text_cost=input_text_cost,
            input_audio_cost=input_audio_cost,
            input_image_cost=input_image_cost,

            cached_text_cost=cached_text_cost,
            cached_audio_cost=cached_audio_cost,
            cached_image_cost=cached_image_cost,

            text_output_cost=text_output_cost,
            audio_output_cost=audio_output_cost,

            total_input_cost=total_input_cost,
            total_output_cost=total_output_cost,
            total_cost=total_cost,
        )


class PricingTextCompletion(BasePricing):
    def compute_costs(self, u: ResponseUsage) -> ChatCompletionModel:
        input_text_tokens = u.input_tokens
        cached_text_tokens = u.input_tokens_details.cached_tokens
        output_text_tokens = u.output_tokens

        input_text_cost = (
            input_text_tokens * self.pricing.get("text_in", 0)
        ) / TOKENS_PER_PRICE

        cached_text_cost = (
            cached_text_tokens * self.pricing.get("cached_text_in", 0)
        ) / TOKENS_PER_PRICE

        output_text_cost = (
            output_text_tokens * self.pricing.get("text_out", 0)
        ) / TOKENS_PER_PRICE

        total_cost = input_text_cost + cached_text_cost + output_text_cost

        return ChatCompletionModel(
            input_text_tokens=input_text_tokens,
            cached_text_tokens=cached_text_tokens,
            output_text_tokens=output_text_tokens,
            input_text_cost=input_text_cost,
            cached_text_cost=cached_text_cost,
            output_text_cost=output_text_cost,
            total_cost=total_cost,
        )


class PricingAudioTranscription(BasePricing):
    def compute_costs(self, u: UsageTranscriptTextUsageTokens) -> RealtimeTranscribeModel:
        input_text_tokens = u.input_token_details.text_tokens
        input_audio_tokens = u.input_token_details.audio_tokens
        output_text_tokens = u.output_tokens

        input_text_cost = (
            input_text_tokens * self.pricing.get("text_in", 0)
        ) / TOKENS_PER_PRICE

        input_audio_cost = (
            input_audio_tokens * self.pricing.get("audio_in", 0)
        ) / TOKENS_PER_PRICE

        output_text_cost = (
            output_text_tokens * self.pricing.get("text_out", 0)
        ) / TOKENS_PER_PRICE

        total_cost = input_text_cost + input_audio_cost + output_text_cost

        return RealtimeTranscribeModel(
            input_text_tokens=input_text_tokens,
            input_audio_tokens=input_audio_tokens,
            output_text_tokens=output_text_tokens,
            input_text_cost=input_text_cost,
            input_audio_cost=input_audio_cost,
            output_text_cost=output_text_cost,
            total_cost=total_cost,
        )



import logging
from pathlib import Path
from pydantic import BaseModel
from .types import UsageTypes
from .utils import get_fields_for_model


class LogWriter:
    """Log writer that logs to a Python logger and computes cost data."""

    def __init__(self, name: str | Path):
        self.logger = logging.getLogger(name)

        if not self.logger.handlers:
            # Set up the logger format and console stream handler if not already set
            self.logger.setLevel(logging.INFO)
            log_format = (
                "%(asctime)s | %(message)s"
            )
            formatter = logging.Formatter(log_format)
            handler = logging.StreamHandler()  # You can change this to a FileHandler if needed
            handler.setFormatter(formatter)
            self.logger.addHandler(handler)

    def initialize(self, pricing: dict[str, float]) -> None: ...

    def write(
            self,
            model: str,
            result: str,
            usage: UsageTypes,
            costs: BaseModel,
    ):
        cost_parts = []
        for label, value in get_fields_for_model(costs).items():
            if isinstance(value, float):
                cost_parts.append(f"{label}: {value:.6f}")
            else:
                cost_parts.append(f"{label}: {value}")

        log_message = (
                f"Model: {model} | "
                f"Result: {result} | "
                f"Total Tokens: {usage.total_tokens} | "
                f"Input Tokens: {usage.input_tokens} | "
                f"Output Tokens: {usage.output_tokens} | "
                + " | ".join(cost_parts)
        )

        self.logger.info(log_message)

    def close(self):
        # No resources to close, logging just happens to the stream or file handler
        pass



import pytest
import csv
from gpt_token_tracker.pricing import TOKENS_PER_PRICE


# -----------------------------
# Realtime CSV tests
# -----------------------------

def test_csv_writer_realtime(
    csv_token_logger_realtime,
    fake_realtime_usage,
    realtime_model,
    realtime_result,
    csv_writer_path,
):

    csv_token_logger_realtime.record(
        realtime_model,
        realtime_result,
        fake_realtime_usage,
    )

    with csv_writer_path.open() as f:
        header_row = f.readline()
        value_row = f.readline()

    assert header_row == "Model,Result,Total Tokens,Input Tokens,Output Tokens,Input Text Tokens,Input Audio Tokens,Input Image Tokens,Cached Text Tokens,Cached Audio Tokens,Cached Image Tokens,Output Text Tokens,Output Audio Tokens,Input Text Cost,Input Audio Cost,Input Image Cost,Cached Text Cost,Cached Audio Cost,Cached Image Cost,Text Output Cost,Audio Output Cost,Total Input Cost,Total Output Cost,Total Cost\n"
    assert value_row == 'gpt-realtime-mini,The capital of Ireland is Dublin,215,165,50,50,100,10,5,0,0,30,20,3e-05,0.001,5e-05,3e-07,0.0,0.0,7.2e-05,0.0004,0.0010803,0.00047200000000000003,0.0015523\n'


# -----------------------------
# Text realtime transcribe CSV tests
# -----------------------------


def test_csv_writer_transcribe_realtime(
    csv_token_logger_realtime_transcribe,
    fake_audio_transcription_usage,
    transcribe_model,
    realtime_transcription_result,
    csv_writer_path,
):

    csv_token_logger_realtime_transcribe.record(
        transcribe_model,
        realtime_transcription_result,
        fake_audio_transcription_usage,
    )

    with csv_writer_path.open() as f:
        header_row = f.readline()
        value_row = f.readline()

    assert header_row == 'Model,Result,Total Tokens,Input Tokens,Output Tokens,Input Text Tokens,Input Audio Tokens,Output Text Tokens,Input Text Cost,Input Audio Cost,Output Text Cost,Total Cost\n'
    assert value_row == 'gpt-4o-mini-transcribe,What is the capital of Ireland?,109,102,7,1,101,7,6e-07,0.00101,1.6800000000000002e-05,0.0010274\n'

# -----------------------------
# Text completion CSV tests
# -----------------------------

def test_csv_writer_text_completion(
    csv_token_logger_text_completion,
    fake_text_completion_usage,
    text_completion_model,
    text_completion_result,
    csv_writer_path,
):

    csv_token_logger_text_completion.record(
        text_completion_model,
        text_completion_result,
        fake_text_completion_usage,
    )

    with csv_writer_path.open() as f:
        header_row = f.readline()
        value_row = f.readline()

    assert header_row == "Model,Result,Total Tokens,Input Tokens,Output Tokens,Input Text Tokens,Cached Text Tokens,Output Text Tokens,Input Text Cost,Cached Text Cost,Output Text Cost,Total Cost\n"
    assert value_row == "gpt-5-mini,The capital of Ireland is Dublin,215,165,50,165,5,50,4.125e-05,1.25e-07,0.0001,0.000141375\n"


# -----------------------------
# Append behavior
# -----------------------------

def test_csv_writer_appends_rows(
    csv_token_logger_text_completion,
    fake_text_completion_usage,
    text_completion_model,
    text_completion_result,
    csv_writer_path,
):

    csv_token_logger_text_completion.record(
        text_completion_model,
        text_completion_result,
        fake_text_completion_usage,
    )
    csv_token_logger_text_completion.record(
        text_completion_model,
        text_completion_result,
        fake_text_completion_usage,
    )

    with csv_writer_path.open() as f:
        rows = list(csv.DictReader(f))

    assert len(rows) == 2



import csv
from pathlib import Path
from pydantic import BaseModel
from .utils import get_fields_for_model
from .types import UsageTypes


class CSVWriter:
    """CSV writer that logs usage and cost models in a structured format."""

    def __init__(self, name: str | Path):
        self.path = Path(name)
        self.path.parent.mkdir(parents=True, exist_ok=True)

        file_exists = self.path.exists()
        self._file = self.path.open("a", newline="")
        self._writer: csv.DictWriter | None = None
        self._header_written = file_exists

    def initialize(self, pricing: dict[str, float]) -> None:
        # CSV does not need pricing, but keep API parity with other writers
        pass

    def write(
        self,
        model: str,
        result: str,
        usage: UsageTypes,
        costs: BaseModel,
    ):
        row = {
            "Model": model,
            "Result": result,
            "Total Tokens": usage.total_tokens,
            "Input Tokens": usage.input_tokens,
            "Output Tokens": usage.output_tokens,
        }

        row.update(get_fields_for_model(costs))

        if self._writer is None:
            self._writer = csv.DictWriter(self._file, fieldnames=row.keys())

            if not self._header_written:
                self._writer.writeheader()
                self._header_written = True

        self._writer.writerow(row)
        self._file.flush()

    def close(self):
        if self._file:
            self._file.close()
            self._file = None
            self._writer = None
